from django import template
from ..models import Link
from django.conf import settings
from django.core.cache import cache


register = template.Library()


@register.inclusion_tag('referrals/referral_token.html', takes_context=True)
def token(context):
	address = settings.DJANGO_REFERRALS_FORM_URL

	if context['request'].user.is_authenticated:
		user = context['request'].user

		if '{}_referral_link'.format(user) in cache:
			token = cache.get('{}_referral_link'.format(user)) 
		else:
			link, created = Link.objects.get_or_create(user=user)
			token = link.token
			cache.set(
				'{}_referral_link'.format(user),
				'{}'.format(token),
				60*60*24
			)
		return {
			'link': '{}?ref={}'.format(address, token)
		}
	
	default_token = settings.DJANGO_REFERRALS_DEFAULT_INPUT_VALUE
	return {
		'link': '{}?ref={}'.format(address, default_token)
	}


#from django.contrib.auth import get_user_model
from django.db.models.signals import post_save
from django.dispatch import receiver
import django.dispatch
import logging
from .models import Link, FlatReferral, MultiLevelReferral
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError


logger = logging.getLogger(__name__)


create_flat_referral = django.dispatch.Signal(providing_args=["request", "user"])
create_multi_level_referral = django.dispatch.Signal(providing_args=["request", "user", "position"])



@receiver(create_flat_referral, sender=None)
def save_flat_referral(sender, request, user, **kwargs):

	referral_link = request.POST.get('referral', None)
	if not referral_link:
		referral_link = request.GET.get('ref', None)

	try:
		if referral_link:
			link = Link.objects.get(token = referral_link)
			referral = FlatReferral.objects.create(
				referrer = link.user,
				referred = user
			)
	except Link.DoesNotExist: # pragma: no cover
		logger.exception('Link with token = {} does not exist'.format(referral_link))
	except ValidationError: # pragma: no cover
		logger.exception('{} is not a valid Link.token'.format(referral_link))


'''
@receiver(post_save, sender=get_user_model())
def create_link(sender, instance. created, **kwargs):
	if created:
		Link.objects.create(user=instance)

'''

@receiver(create_multi_level_referral, sender=None)
def save_multi_level_referral(sender, request, user, position, **kwargs):
	if position is not 'child' and position is not 'sibling':
		logger.exception('{} was sent instead of a child or brother'.format(position)) # pragma: no cover
		raise KeyError('Position can be child or sibling') # pragma: no cover
	
	referral_link = request.POST.get('referral', None)
	if not referral_link:
		referral_link = request.GET.get('ref', None)
	try: 
		if referral_link:
			link = Link.objects.get(token = referral_link)
			try:
				referral = MultiLevelReferral.objects.get(user = link.user)
				if position == 'sibling':
					referral.add_sibling(user = user)
				else:
					referral.add_child(user = user)
			except MultiLevelReferral.DoesNotExist: # pragma: no cover
				logger.exception('MultiLevelReferral with user = {} does not exist'.format(link.user))
	except Link.DoesNotExist: # pragma: no cover
		logger.exception('Link with token = {} does not exist'.format(referral_link))
	except ValidationError: # pragma: no cover
		logger.exception('{} is not a valid Link.token'.format(referral_link))


# -*- coding: utf-8
from django.apps import AppConfig


class ReferralsConfig(AppConfig):
	name = 'referrals'
	verbose_name = 'Referrals'

	def ready(self):
		"""Override this to put in:
			Referrals system checks
			Referrals signal registration
		"""
		try: # pragma: no cover
			import referrals.signals  # noqa F401
		except ImportError: # pragma: no cover
			pass


from django.contrib.auth.mixins import LoginRequiredMixin
from django.urls import reverse
from django.views.generic import DetailView, ListView, RedirectView, UpdateView
from .models import FlatReferral


class FlatReferralDetailView(LoginRequiredMixin, DetailView):

	model = FlatReferral
	template_name = 'referrals/flat_referral_detail.html'


class FlatReferralListView(LoginRequiredMixin, ListView):

	model = FlatReferral
	template_name = 'referrals/flat_referral_list.html'
	context_object_name = 'flat_referral_list'
	paginate_by = 10

	def get_queryset(self):
		queryset = super(FlatReferralListView, self).get_queryset()
		return queryset.filter(referrer=self.request.user).all()


# -*- coding: utf-8 -*-
from django.conf.urls import url
from django.views.generic import TemplateView

from . import views

app_name = 'referrals'
urlpatterns = [
    #url(r'', TemplateView.as_view(template_name="base.html")),
    url(
    	r'^$',
    	view = views.FlatReferralListView.as_view(),
    	name = 'flat_referral_list'
    ),
    url(
    	r'^(?P<pk>\d+)/$',
    	view = views.FlatReferralDetailView.as_view(),
    	name = 'flat_referral_detail'
    )
]



from django.contrib import admin
from treebeard.admin import TreeAdmin
from treebeard.forms import movenodeform_factory
from .models import FlatReferral, Link, MultiLevelReferral

class MultiLevelReferralAdmin(TreeAdmin):
	form = movenodeform_factory(MultiLevelReferral)

admin.site.register(MultiLevelReferral, MultiLevelReferralAdmin)


admin.site.register(FlatReferral)
admin.site.register(Link)



# Generated by Django 2.0.2 on 2018-04-15 11:42

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='FlatReferral',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('referred', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='referrers', to=settings.AUTH_USER_MODEL)),
                ('referrer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='referreds', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'flat_referral',
                'ordering': ['id'],
            },
        ),
        migrations.CreateModel(
            name='Link',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('token', models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='MultiLevelReferral',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('path', models.CharField(max_length=255, unique=True)),
                ('depth', models.PositiveIntegerField()),
                ('numchild', models.PositiveIntegerField(default=0)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'db_table': 'multi_level_referral',
            },
        ),
        migrations.AlterUniqueTogether(
            name='flatreferral',
            unique_together={('referrer', 'referred')},
        ),
    ]



from django.forms.widgets import Widget
from django.template import loader
from django.utils.safestring import mark_safe
from django.conf import settings


class ReferralWidget(Widget):
	template_name = 'referrals/referral_widget_min.html'

	def get_context(self, name, value, attrs=None):
		return {
			'widget': {
				'name': name,
				'value': value,
				'default_value': settings.DJANGO_REFERRALS_DEFAULT_INPUT_VALUE
			}
		}

	def render(self, name, value, renderer=None, attrs=None):
		context = self.get_context(name, value, attrs)
		template = loader.get_template(self.template_name).render(context)
		return mark_safe(template)


# -*- coding: utf-8 -*-

from django.db import models
from django.conf import settings
from django.utils.encoding import python_2_unicode_compatible
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext_lazy as _
from django.urls import reverse
from django.contrib.auth import get_user_model
import uuid
from treebeard.mp_tree import MP_Node


@python_2_unicode_compatible
class FlatReferral(models.Model):

	referrer = models.ForeignKey(
		get_user_model(), 
		on_delete = models.CASCADE,
		related_name = 'referreds'
	)
	referred = models.OneToOneField(
		get_user_model(),
		on_delete = models.CASCADE,
		related_name = 'referrers'
	)

	class Meta:
		unique_together = (('referrer', 'referred'),)
		ordering = ['id']
		db_table = 'flat_referral'
	
	def __str__(self):
		return '{} => {}'.format(self.referrer, self.referred)

	def clean(self, *args, **kwargs):
		if self.referrer == self.referred:
			raise ValidationError(_('The referrer can not be referred.'))


	def save(self, *args, **kwargs):
		self.full_clean()
		return super(FlatReferral, self).save(*args, **kwargs)

	def get_absolute_url(self):
		return reverse('referrals:flat_referral_detail', kwargs={'pk': self.pk})


class Link(models.Model):

	user = models.OneToOneField(
		get_user_model(),
		on_delete = models.CASCADE
	)

	token = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)


@python_2_unicode_compatible
class MultiLevelReferral(MP_Node):
	user = models.OneToOneField(
		get_user_model(),
		on_delete = models.CASCADE
	)

	node_order_by = ['user']

	class Meta:
		db_table = 'multi_level_referral'


	def __str__(self):
		return self.user.username